using System;
using System.Collections;
using System.Drawing;
using System.Windows.Forms;

using RSSReader.OPML;
using RSSReader.RSS;

namespace RSSReader.Controls
{

	#region PersistedTreeState class 

	public class PersistedTreeState
	{

		#region Private member variables 

		private Hashtable	states;
		private string		selectedPath;
		private TreeView	tree;

		#endregion 

		#region Constructor 

		private PersistedTreeState() {}

		public PersistedTreeState( TreeView Tree )
		{

			tree = Tree;
			states = SaveTreeState();

			if( tree.SelectedNode != null )
				selectedPath = tree.SelectedNode.FullPath;
			else
				selectedPath = "";

		}

		#endregion 

		#region Public methods 

		public void Restore()
		{

			this.RestoreTreeState( states );

			if( selectedPath != "" )
			{
			
				TreeNode selected = TreeViewHelper.FindNodeByPath( tree, selectedPath );
				if( selected != null )
				{
					tree.SelectedNode = selected;
					selected.EnsureVisible();
				}

			}

		}

		#endregion 

		#region Private utility methods 

		private Hashtable SaveTreeState()
		{

			Hashtable states = new Hashtable();
			SaveNodeState( states, tree.Nodes );

			return states;

		}

		private void SaveNodeState( Hashtable states, TreeNodeCollection nodes )
		{

			foreach( TreeNode node in nodes )
			{
				states[ node.FullPath ] = node.IsExpanded;
				SaveNodeState( states, node.Nodes );
			}

		}

		private void RestoreTreeState( Hashtable states )
		{
			RestoreNodeState( states, tree.Nodes );
		}

		private void RestoreNodeState( Hashtable states, TreeNodeCollection nodes )
		{

			foreach( TreeNode node in nodes )
			{

				if( !states.ContainsKey( node.FullPath ) )
					continue;

				bool isExpanded = (bool) states[ node.FullPath ];
				if( isExpanded )
					node.Expand();


				RestoreNodeState( states, node.Nodes );

			}

		}

		#endregion 

	}

	#endregion 

	#region TreeViewHelper class 

	/// <summary>
	/// Contains common routines for handling TreeView objects
	/// </summary>
	public class TreeViewHelper
	{

		#region Public methods 

		public static TreeNode FindNodeByPath( TreeView Tree, string FullPath )
		{
			return FindNodeByPath( Tree.Nodes, FullPath );
		}

		public static ArrayList GetAllNodes( TreeView Tree )
		{

			ArrayList nodes = new ArrayList();
			GetAllNodes( nodes, Tree.Nodes );

			return nodes;

		}

		#endregion 

		#region Private utility methods 

		private static void GetAllNodes( ArrayList list, TreeNodeCollection nodes )
		{

			foreach( TreeNode node in nodes )
			{
				list.Add( node );
				GetAllNodes( list, node.Nodes );
			}

		}

		private static TreeNode FindNodeByPath( TreeNodeCollection Nodes, string FullPath )
		{

			foreach( TreeNode node in Nodes )
			{

				if( node.FullPath.ToLower() == FullPath.ToLower() )
					return node;

				TreeNode search = FindNodeByPath( node.Nodes, FullPath );
				if( search != null )
					return search;

			}

			return null;

		}

		#endregion 

	}

	#endregion 

	#region NodeWithSummary base class 

	public class NodeWithSummary : TreeNode, IOwnerDrawNode
	{

		#region Public properties 

		public virtual int UnreadMessageCount
		{
		
			get
			{

				int unreadMessages = 0;
				foreach( TreeNode node in this.Nodes )
				{

					if( node is NodeWithSummary )
						unreadMessages += ((NodeWithSummary) node).UnreadMessageCount;

				}

				return unreadMessages;

			}

		}

		#endregion 

		#region Public methods 

		public virtual void RefreshNode()
		{

			if( this.Parent is NodeWithSummary )
				((NodeWithSummary) this.Parent).RefreshNode();
			
			if( this.IsVisible )
				this.TreeView.Refresh();

		}

		#endregion 

		#region IOwnerDrawNode Members 

		private Font GetRenderFont()
		{

			Font nodeFont = SubscriptionExplorer.NodeDefaultFont;
			if( UnreadMessageCount > 0 )
				nodeFont = SubscriptionExplorer.NodeBoldFont;

			return nodeFont;

		}

		public void OnPaint(bool IsSelected, PaintEventArgs e)
		{
			
			if( this.IsEditing )
				return;

			e.Graphics.SmoothingMode = System.Drawing.Drawing2D.SmoothingMode.None;
			
			int unreadMessages = this.UnreadMessageCount;

			Font nodeFont = GetRenderFont();

			// Should fix at least some of the bounding-box problems...
			this.NodeFont = nodeFont;

			SizeF textSize = e.Graphics.MeasureString( Text, nodeFont, Screen.PrimaryScreen.WorkingArea.Width );

			Color foreColor = (this.ForeColor != Color.Empty) ? this.ForeColor : TreeView.ForeColor;
			Color backColor = (this.BackColor != Color.Empty) ? this.BackColor : TreeView.BackColor;
			Color hiColor = Color.Blue;
			if( IsSelected )
			{

				if( TreeView.Focused )
				{
					backColor = SystemColors.Highlight;
					foreColor = SystemColors.HighlightText;
					hiColor = SystemColors.HighlightText;
				}
				else
				{
					backColor = SystemColors.ControlLight;
					foreColor = SystemColors.ControlText;
				}

			}

			using( SolidBrush backBrush = new SolidBrush( backColor ) )
			{
				e.Graphics.FillRectangle( backBrush, NodeBounds );
			}

			int textTop = Bounds.Top + (int)((Bounds.Height - textSize.Height) / 2);

			using( SolidBrush foreground = new SolidBrush( foreColor ) )
			{

				e.Graphics.DrawString( 
					Text, 
					nodeFont, 
					foreground,
					Bounds.Left,
					textTop
					);

			}

			if( unreadMessages > 0 )
			{

				using( SolidBrush hi = new SolidBrush( hiColor ) )
				{

					e.Graphics.DrawString( 
						"(" + unreadMessages + ")", 
						SubscriptionExplorer.NodeDefaultFont, 
						hi, 
						Bounds.Left + textSize.Width + 3, 
						textTop
						);

				}

			}

		}

		private int MeasureTextWidth()
		{

			using( Graphics g = TreeView.CreateGraphics() )
			{

				Font font = this.GetRenderFont();
				string text = this.Text;
				if( this.UnreadMessageCount > 0 )
					text += " (" + this.UnreadMessageCount + ")";

				// We just *really* want to make sure that the MeasureString() call 
				// does not attempt to word-wrap the text.  There's almost certainly
				// a better way to do this, but it works.
				int width = Math.Max( TreeView.Width, TreeView.ClientRectangle.Width ) * 2;

				return (int) g.MeasureString( text, font, width ).Width + 15;

			}

		}

		public Rectangle NodeBounds
		{

			get
			{

				int textWidth = MeasureTextWidth();

				Rectangle bounds = new Rectangle(
					base.Bounds.Left,
					base.Bounds.Top,
					Math.Max( base.Bounds.Width, textWidth ),
					base.Bounds.Height
					);

				return bounds;

			}

		}

		#endregion

	}

	#endregion 

	#region OPMLTreeNode base class 

	/// <summary>
	/// This class is mostly just a marker interface so that we know whether a given
	/// node in the tree represents an OPML file element, but also provides a class
	/// factory for creating TreeNode objects based on an OPML file element.
	/// </summary>
	public class OPMLTreeNode : NodeWithSummary
	{

		public static OPMLTreeNode Create( OPMLNode SubscriptionNode )
		{

			if( SubscriptionNode is OPMLCategoryNode )
				return new CategoryTreeNode( (OPMLCategoryNode) SubscriptionNode );
			else
				return new SubscriptionTreeNode( (OPMLItemNode) SubscriptionNode );

		}

	}

	#endregion 

	#region CategoryTreeNode class 

	public class CategoryTreeNode : OPMLTreeNode
	{

		#region Private member variables 

		private OPMLCategoryNode	category;

		#endregion 

		#region Constructor 

		public CategoryTreeNode( OPMLCategoryNode Category )
		{
			
			this.category = Category;
			this.Tag = Category;

			this.Text = Category.Title;

		}

		#endregion 

		#region Public properties 

		public OPMLCategoryNode Category
		{
			get { return category; }
			set { category = value; }
		}

		#endregion 

	}

	#endregion 

	#region SubscriptionTreeNode class 

	public class SubscriptionTreeNode : OPMLTreeNode
	{

		#region Private data members 

		private OPMLItemNode	subscription;
		private int				unreadMessageCount = int.MinValue;

		#endregion 

		#region Constructor 

		public SubscriptionTreeNode( OPMLItemNode Subscription )
		{

			this.subscription = Subscription;
			this.Tag = Subscription;

			this.Text = System.Web.HttpUtility.HtmlDecode( Subscription.Title );

			RSSFeed feed = FeedManager.GetCachedFeed( Subscription.XmlUrl );
			if( feed != null )
				RefreshNode( feed );
			else
				FeedManager.RetrieveFeedAsync( subscription.XmlUrl, "first", new FeedManager.RetrieveFeedAsyncCallback( OnFeedRetrieved ) );
					
		}

		#endregion 

		#region Public properties 

		public OPMLItemNode Subscription
		{
			
			get { return subscription; }
			
			set 
			{ 
				subscription = value; 
				unreadMessageCount = UnreadMessageCount;
			}

		}

		public override int UnreadMessageCount
		{

			get
			{

				if( unreadMessageCount > int.MinValue )
					return unreadMessageCount;

				RSSFeed feed = FeedManager.GetCachedFeed( subscription.XmlUrl );
				if( feed == null )
					return 0;
				else
					return feed.GetItemCount() - feed.GetItemsReadCount();

			}

		}

		public bool HasFollowUpItems
		{

			get
			{

				RSSFeed feed = FeedManager.GetCachedFeed( subscription.XmlUrl );
				if( feed == null )
					return false;
			
				foreach( RSSItem item in feed.Items )
					if( item.MarkedForFollowUp )
						return true;

				return false;

			}

		}

		#endregion 

		#region Public methods 

		public void RefreshNode( RSSFeed feed )
		{

			this.Text = subscription.Title;
			unreadMessageCount = feed.GetItemCount() - feed.GetItemsReadCount();

			base.RefreshNode();

		}

		#endregion 

		#region Private utility methods 

		private void OnFeedRetrieved( RSSFeed feed, object Tag )
		{

			RefreshNode( feed );

			// If this is the first time the feed has been retrieved, save it to 
			// the local cache.
			if( (Tag is string) && ((string) Tag) == "first" )
				FeedManager.SaveFeed( feed );

		}

		#endregion 

	}

	#endregion 

	#region MyFeedsNode class 

	public class MyFeedsNode : NodeWithSummary
	{
	}

	#endregion 

	#region FollowUpItemsNode class 

	public class FollowUpItemsNode : NodeWithSummary
	{
	}

	#endregion 

}
