using System;
using System.IO;
using System.Text;
using System.Collections;

namespace RSSReader
{

	internal class StringTokenizer
	{

		#region Private data members 

		private const char EOF = (char)0;

		private int line;
		private int column;
		private int currentPos;

		private string data;

		private bool	ignoreWhiteSpace;
		private char[]	symbolChars;
		private char	escapeChar;

		private int saveLine;
		private int saveCol;
		private int savePos;

		#endregion 

		#region Constructor 

		public StringTokenizer()
		{
			this.escapeChar = '\\';
			this.ignoreWhiteSpace = false;
			this.symbolChars = new char[] {'=', '+', '-', '/', ',', '.', '*', '~', '!', '@', '#', '$', '%', '^', '&', '(', ')', '{', '}', '[', ']', ':', ';', '<', '>', '?', '|', '\\'};
		}

		#endregion 

		#region Public properties 

		/// <summary>
		/// gets or sets which characters are part of TokenKind.Symbol
		/// </summary>
		public char[] SymbolChars
		{
			get { return this.symbolChars; }
			set { this.symbolChars = value; }
		}

		/// <summary>
		/// if set to true, white space characters will be ignored,
		/// but EOL and whitespace inside of string will still be tokenized
		/// </summary>
		public bool IgnoreWhiteSpace
		{
			get { return this.ignoreWhiteSpace; }
			set { this.ignoreWhiteSpace = value; }
		}

		#endregion 

		#region Private utility methods 

		protected char LookAhead(int count)
		{

			if(currentPos + count >= data.Length)
				return EOF;
			else
				return data[currentPos+count];

		}

		protected char Consume()
		{

			char ret = data[currentPos];
			currentPos++;
			column++;

			return ret;

		}

		#endregion 

		#region Token handling 

		protected Token CreateToken(TokenKind kind, string value)
		{
			return new Token(kind, value, line, column);
		}

		protected Token CreateToken(TokenKind kind)
		{
			string tokenData = data.Substring(savePos, currentPos-savePos);
			return new Token(kind, tokenData, saveLine, saveCol);
		}

		/// <summary>
		/// save read point positions so that CreateToken can use those
		/// </summary>
		private void StartRead()
		{
			saveLine = line;
			saveCol = column;
			savePos = currentPos;
		}

		/// <summary>
		/// reads all whitespace characters (does not include newline)
		/// </summary>
		/// <returns></returns>
		protected Token ReadWhitespace()
		{

			StartRead();

			while( Char.IsWhiteSpace(LookAhead(0)) && LookAhead(0) != '\r' && LookAhead(0) != '\n' )
				Consume();

			return CreateToken(TokenKind.WhiteSpace);
			
		}

		/// <summary>
		/// reads number. Number is: DIGIT+ ("." DIGIT*)?
		/// </summary>
		/// <returns></returns>
		protected Token ReadNumber()
		{
			StartRead();

			bool hadDot = false;

			Consume(); // read first digit

			while (true)
			{
				char ch = LookAhead(0);
				if(Char.IsDigit(ch))
					Consume();
				else if(ch == '.' && !hadDot)
				{
					hadDot = true;
					Consume();
				}
				else
					break;
			}

			return CreateToken(TokenKind.Number);
		}

		/// <summary>
		/// reads word. Word contains any alpha character or _
		/// </summary>
		protected Token ReadWord()
		{

			StartRead();

			while( Char.IsLetterOrDigit( LookAhead(0) ) || LookAhead(0) == '_' )
			{
				Consume();
			}
				
			return CreateToken(TokenKind.Word);

		}

		/// <summary>
		/// reads all characters until next " is found.
		/// If "" (2 quotes) are found, then they are consumed as
		/// part of the string
		/// </summary>
		/// <returns></returns>
		protected Token ReadString()
		{
			
			StartRead();

			char delim = LookAhead(0);

			Consume();

			while( true )
			{

				char ch = LookAhead(0);
				if( ch == EOF )
					throw new EndOfStreamException( "Unexpected end of file" );

				if( ch == '\r' || ch == '\n' )
					break;

				if( ch == delim )
				{
					Consume();
					break;
				}
				else if( ch == escapeChar )
				{
					Consume(); // Consume escape character
					Consume(); // Consume escapED character
					continue;
				}
					
				Consume();

			}

			return CreateToken(TokenKind.QuotedString);

		}

		/// <summary>
		/// checks whether c is a symbol character.
		/// </summary>
		protected bool IsSymbol(char c)
		{

			if( Char.IsSymbol( c ) )
				return true;

			for (int i=0; i<symbolChars.Length; i++)
				if(symbolChars[i] == c)
					return true;

			return false;

		}
	
		private Token Next()
		{

			ReadToken:

				char ch = LookAhead(0);
			switch (ch)
			{

				case EOF:
					return CreateToken(TokenKind.EOF, string.Empty);

				case '0':
				case '1':
				case '2':
				case '3':
				case '4':
				case '5':
				case '6':
				case '7':
				case '8':
				case '9':
					return ReadNumber();

				case '\r':
				{

					StartRead();
					Consume();
					if(LookAhead(0) == '\n')
						Consume();	// on DOS/Windows we have \r\n for new line

					line++;
					column=1;

					return CreateToken(TokenKind.EOL);

				}

				case '\n':
				{

					StartRead();
					Consume();
					line++;
					column=1;
					
					return CreateToken(TokenKind.EOL);

				}

				case '"':
				case '\'':
					return ReadString();
					break;

				default:
				{

					if(Char.IsLetter(ch) || ch == '_')
						return ReadWord();
					else if(Char.IsWhiteSpace(ch))
					{
					
						if(this.ignoreWhiteSpace)
						{
							Consume();
							goto ReadToken;
						}
						else
							return ReadWhitespace();

					}
					else if(IsSymbol(ch))
					{
						StartRead();
						Consume();
						return CreateToken(TokenKind.Symbol);
					}
					else
					{
						StartRead();
						Consume();
						return CreateToken(TokenKind.Unknown);						
					}

				}

			}

		}

		#endregion 

		#region Public methods 

		public ArrayList Tokenize( string data )
		{

			ArrayList tokens = new ArrayList();

			this.data = data;
			this.line = 1;
			this.column = 1;
			this.currentPos = 0;

			Token lastToken = null;
			Token lastSignificant = null;

			while( true )
			{

				Token token = this.Next();
				if( token.Kind != TokenKind.EOF )
					tokens.Add( token );
				else
					break;

				if( lastToken != null )
				{
					token.Previous = lastToken;
					lastToken.Next = token;
				}

				if( lastSignificant != null )
					token.LastSignificant = lastSignificant;

				if( token.Kind != TokenKind.WhiteSpace && token.Kind != TokenKind.EOL )
				{

					if( lastToken != null )
						lastToken.NextSignificant = token;

					lastSignificant = token;

				}

				lastToken = token;

			}

			return tokens;

		}

		#endregion 

	}

	#region TokenKind enumeration 

	public enum TokenKind
	{
		Unknown,
		Word,
		Number,
		QuotedString,
		WhiteSpace,
		Symbol,
		EOL,
		EOF
	}

	#endregion 

	#region Token class 

	internal class Token
	{

		#region Private data members 

		private int			line;
		private int			column;
		private string		data;
		private TokenKind	kind;

		private Token		next = null;
		private Token		prev = null;
		private Token		lastSignificant = null;
		private Token		nextSignificant = null;

		#endregion 

		#region Constructor 

		public Token(TokenKind kind, string value, int line, int column)
		{
			this.kind = kind;
			this.data = value;
			this.line = line;
			this.column = column;
		}

		#endregion 

		#region Public properties 

		public int Column
		{
			get { return this.column; }
		}

		public TokenKind Kind
		{
			get { return this.kind; }
		}

		public int Line
		{
			get { return this.line; }
		}

		public string Value
		{
			get { return this.data; }
			set { this.data = value; }
		}

		#endregion 

		#region List members 

		public Token Next
		{
			get { return next; }
			set { next = value; }
		}

		public Token NextSignificant
		{
			get { return nextSignificant; }
			set { nextSignificant = value; }
		}

		public Token Previous
		{
			get { return prev; }
			set { prev = value; }
		}

		public Token LastSignificant
		{
			get { return this.lastSignificant; }
			set { lastSignificant = value; }
		}

		#endregion 

	}

	#endregion 

}
