using System;
using System.IO;
using System.Net;
using System.Text;
using System.Text.RegularExpressions;
using System.Web;
using System.Xml;
using System.Collections;
using System.Collections.Specialized;
using System.Runtime;
using System.Runtime.InteropServices;
using System.Windows.Forms;

namespace RSSReader.RSS
{

	/// <summary>
	/// Searches a given web page (indicated by Url) for links to RSS feeds.
	/// </summary>
	public class RSSFeedLocator
	{

		#region Public events 

		public delegate void FeedbackEventHandler( string Url );
		public event FeedbackEventHandler Searching;

		public delegate void FeedFoundEventHandler( RSSFeed Feed );
		public event FeedFoundEventHandler FeedFound;

		#endregion 

		#region Private variables 

		private Hashtable	searched;

		#endregion 

		#region Constructor 

		public RSSFeedLocator()
		{
			searched = new Hashtable();
		}

		#endregion 

		#region Public methods 

		public void Search( string url )
		{

			searched = new Hashtable();
			CheckForRSSLinks( url, 1 );

		}

		#endregion

		#region Private utility methods 

		private bool IsRSSContentType( string contentType )
		{

			contentType = contentType.ToLower();

			string[] matchTypes = { "text/xml", "application/xml", "text/rss", "xml", "rss" };
			foreach( string matchType in matchTypes )
				if( contentType.StartsWith( matchType ) )
					return true;

			return false;

		}

		private ArrayList ExtractLinksFromHtml( string htmlData )
		{

			ArrayList links = new ArrayList();

			// Patterns that might work:
			// @"(http|ftp|https):\/\/[\w-_]+(\.[\w-_]+)+([\w\-\.,@?^=%&:/~\+#]*[\w\-\@?^=%&/~\+#])?"
			// "(?<http>(http:[/][/]|www.)([a-z]|[A-Z]|[0-9]|[/.]|[~])*)"

			Regex pattern = new Regex( @"(http|https):\/\/[\w-_]+(\.[\w-_]+)+([\w\-\.,@?^=%&:/~\+#]*[\w\-\@?^=%&/~\+#])?", RegexOptions.IgnoreCase | RegexOptions.Multiline );
			foreach( Match match in pattern.Matches( htmlData ) )
			{
				links.Add( match.Value );						
			}

			return links;

		}

		private void CheckForRSSLinks( string url, int crawlLevel )
		{

			try
			{

				// We do not want to allow the user to subscribe to a feed
				// more than once, and if a feed is already cached then 
				// we know that the user is already subscribed.
				if( FeedManager.IsFeedCached( url ) )
					return;

				if( searched.ContainsKey( url.ToLower() ) )
					return;
				else
					searched.Add( url.ToLower(), url );

				if( Searching != null )
					Searching( url );

				HttpWebRequest request		= (HttpWebRequest) System.Net.WebRequest.Create( url );
				request.Timeout				= (int) TimeSpan.FromSeconds( 25 ).TotalMilliseconds;
				request.UserAgent			= "SharpTools RSS Reader V0.3A"; 
				request.AllowAutoRedirect	= false;
				request.Referer				= "http://www.ms-inc.net";
				request.KeepAlive			= false;	
				request.Proxy				= GlobalProxySelection.GetEmptyWebProxy();
				request.Proxy.Credentials	= CredentialCache.DefaultCredentials;

				using( HttpWebResponse response = (HttpWebResponse) request.GetResponse() )
				{

					if( response.StatusCode != HttpStatusCode.OK )
						return;

					if( response.ContentLength == 0 )
						return;

					#region Check to see if link goes directly to RSS feed 

					if( IsRSSContentType( response.ContentType ) )
					{

						string xmlData = "";
						using( Stream stream = response.GetResponseStream() )
						{	

							XmlTextReader xmlReader = new XmlTextReader( stream );
							xmlReader.XmlResolver = null;
							if( xmlReader.MoveToContent() != XmlNodeType.Element )
								return;

							if( xmlReader.LocalName == "rss" || xmlReader.LocalName == "RDF" )
							{
							
								XmlDocument doc = new XmlDocument();
								doc.Load( xmlReader );
							
								RSSFeed feed = new RSSFeed();
								feed.ReadFromXml( doc.DocumentElement, url );

								if( FeedFound != null )
									FeedFound( feed );

							}

						}

						return;

					}

					#endregion 

					if( !response.ContentType.ToLower().StartsWith( "text/htm" ) )
						return;

					string htmlData = "";
					using( StreamReader htmlReader = new StreamReader( response.GetResponseStream() ) )
					{
						htmlData = htmlReader.ReadToEnd();
					}

					CheckForRSSLinkElement( htmlData );

					if( crawlLevel == 0 )
						return;

					ArrayList links = ExtractLinksFromHtml( htmlData );
					foreach( string link in links )
						CheckForRSSLinks( link, crawlLevel - 1 );

				}

			}
			catch( Exception err )
			{
				System.Diagnostics.Debug.WriteLine( "Failed to load document: " + err.ToString() );
			}

		}

		private StringDictionary ParseAttributes( string element )
		{

			StringDictionary results = new StringDictionary();

			Regex pattern = new Regex( "(\\s*(\\w*\\s*)=\\s*(\"[^\"]*\"|'[^']'|[^>]))?", RegexOptions.IgnoreCase | RegexOptions.Multiline );
			foreach( Match match in pattern.Matches( element ) )
			{

				string attributeName = match.Groups[2].Value.ToLower();
				if( attributeName == "" )
					continue;

				string attributeValue = match.Groups[3].Value.Trim().Trim( '"', '\'' );

				if( !results.ContainsKey( attributeName) )
					results.Add( attributeName, attributeValue );

			}

			return results;

		}

		private void CheckForRSSLinkElement( string htmlData )
		{

			Regex pattern = new Regex( @"<([^\s>]*)(\s[^<]*)>", RegexOptions.IgnoreCase | RegexOptions.Multiline );
			foreach( Match match in pattern.Matches( htmlData ) )
			{

				if( match.Groups[1].Value.ToLower() != "link" )
					continue;

				StringDictionary attributes = ParseAttributes( match.Groups[2].Value.Trim() );
				if( attributes["type"] == "application/rss+xml" )
				{

					string url = attributes["href"];
					if( url != null && url != "" && !searched.ContainsKey( url.ToLower() ) )
						CheckForRSSLinks( url, 0 );

				}

			}

		}

		#endregion 

	}

}
